# Author: Mark Blakeney, Feb 2024.
# PYTHON_ARGCOMPLETE_OK
"""
Install Python applications into isolated virtual environments and
create links to the executables in a bin directory for your PATH. Like
pipx but uses uv instead of venv + pip.
"""

from __future__ import annotations

import importlib
import os
import platform
import re
import shlex
import sys
from argparse import ArgumentParser
from pathlib import Path

import argcomplete
import platformdirs

from . import utils
from .run import run

DEFUV = 'uv'
MIN_UV_VERSION = '0.1.34'
DEFPY = 'python3'

# Some constants
MOD = Path(__file__)
PROG = MOD.stem
PROGU = PROG.upper()
CNFFILE = platformdirs.user_config_path(f'{PROG}-flags.conf')


def calc_version(verstr: str) -> tuple[int, ...]:
    "Calculate a version tuple from a version string"
    return tuple(int(x) for x in verstr.split('.'))


def path_check(bin_name: str, bin_dir: str) -> str:
    "Check and report that users PATH is set up correctly"
    if not (path := os.getenv('PATH')):
        return 'WARNING: Your PATH is not set.'

    if bin_dir not in path.split(':'):
        return f'WARNING: Your PATH does not contain {bin_name} ({bin_dir}).'

    return f'Your PATH contains {bin_name} ({bin_dir}).'


def main() -> str | None:
    "Main code"
    mainparser = ArgumentParser(
        description=__doc__,
        epilog='Some commands offer aliases as shown in parentheses above. '
        f'Note you can set default starting global options in {CNFFILE}.',
    )
    mainparser.add_argument(
        '--uv', metavar='uv_path', help=f'path to uv executable, default="{DEFUV}"'
    )
    mainparser.add_argument(
        '-m',
        '--no-man-pages',
        action='store_true',
        help='do not install package man pages',
    )
    mainparser.add_argument('--home', help=f'specify {PROGU}_HOME')
    mainparser.add_argument('--bin-dir', help=f'specify {PROGU}_BIN_DIR')
    mainparser.add_argument('--man-dir', help=f'specify {PROGU}_MAN_DIR')
    mainparser.add_argument(
        '--default-python', help=f'path to default python executable, default="{DEFPY}"'
    )
    mainparser.add_argument(
        '-V',
        '--version',
        action='store_true',
        help=f'just print {PROG} version and exit',
    )
    subparser = mainparser.add_subparsers(title='Commands', dest='func')

    # Iterate over the commands to set up their parsers
    for modfile in sorted((MOD.parent / 'commands').glob('[!_]*.py')):
        name = modfile.stem
        mod = importlib.import_module(f'{PROG}.commands.{name}')
        aliases = mod.aliases if hasattr(mod, 'aliases') else []
        docstr = mod.__doc__.strip().split('\n\n')[0] if mod.__doc__ else None
        parser = subparser.add_parser(
            name, description=mod.__doc__, help=docstr, aliases=aliases
        )

        if hasattr(mod, 'init'):
            mod.init(parser)

        if not hasattr(mod, 'main'):
            mainparser.error(f'"{name}" command must define a main()')

        parser.set_defaults(func=mod.main, parser=parser, name=name)

    # Command arguments are now defined, so we can set up argcomplete
    argcomplete.autocomplete(mainparser)

    # Merge in default args from user config file. Then parse the
    # command line.
    if CNFFILE.exists():
        with CNFFILE.open() as fp:
            lines = [re.sub(r'#.*$', '', line).strip() for line in fp]
        cnflines = ' '.join(lines).strip()
    else:
        cnflines = ''

    args = mainparser.parse_args(shlex.split(cnflines) + sys.argv[1:])

    if args.version:
        print(f'{PROG}=={utils.version()}')
        return None

    if platform.system() == 'Windows':
        return 'Error: Sorry, Windows platform is not supported.'

    is_root = os.geteuid() == 0
    home_dir = args.home or os.getenv(f'{PROGU}_HOME')
    bin_dir = args.bin_dir or os.getenv(f'{PROGU}_BIN_DIR')
    man_dir = args.man_dir or os.getenv(f'{PROGU}_MAN_DIR')

    pyexe = utils.subenvars(
        args.default_python
        if args.default_python
        else (os.getenv(f'{PROGU}_DEFAULT_PYTHON') or DEFPY)
    )

    if not home_dir:
        home_dir = (
            f'/opt/{PROG}' if is_root else f'{platformdirs.user_data_dir()}/{PROG}'
        )
    if not bin_dir:
        bin_dir = '/usr/local/bin' if is_root else '~/.local/bin'
    if not man_dir:
        man_dir = '/usr/local/share/man' if is_root else '~/.local/share/man'

    home_dir = utils.subenvars(home_dir, resolve=True)
    bin_dir = utils.subenvars(bin_dir, resolve=True)
    man_dir = utils.subenvars(man_dir, resolve=True)

    if not args.func:
        mainparser.print_help()
        print('\nEnvironment:')
        print(f'{PROGU}_HOME = {home_dir}')
        print(f'{PROGU}_BIN_DIR = {bin_dir}')
        print(f'{PROGU}_MAN_DIR = {man_dir}')
        print(f'{PROGU}_DEFAULT_PYTHON = {pyexe}')
        print()
        print(path_check(f'{PROGU}_BIN_DIR', str(bin_dir)))
        return None

    # Ensure uv is installed/available
    uv = args.uv or DEFUV
    if not (verstr := run((uv, '--version'), capture=True, ignore_error=True)):
        if args.uv:
            return f'Error: specified uv "{uv}" program not found.'

        return (
            f'Error: {uv} program must be installed, and in your PATH '
            'or specified with --uv option.'
        )

    uv_vers = verstr.split()[1]
    if calc_version(uv_vers) < calc_version(MIN_UV_VERSION):
        return (
            f'Error: {uv} version is {uv_vers} but must be at least {MIN_UV_VERSION}.'
        )

    # Keep some useful info in the namespace passed to the command
    args._uv = uv
    args._lockfile = home_dir / f'.{PROG}.lock'
    args._packages_dir = home_dir / 'packages'
    args._packages_dir.mkdir(parents=True, exist_ok=True)
    args._venvs_dir = home_dir / 'venvs'
    args._venvs_dir.mkdir(parents=True, exist_ok=True)
    args._bin_dir = bin_dir
    args._bin_dir.mkdir(parents=True, exist_ok=True)
    args._man_dir = man_dir
    if not args.no_man_pages:
        args._man_dir.mkdir(parents=True, exist_ok=True)
    args._pyexe = pyexe
    args._prog = PROG
    args._meta_file = f'{PROG}_metadata.json'
    args._freeze_file = f'{PROG}_freeze.txt'
    if not hasattr(args, 'verbose'):
        args.verbose = False

    # Purge any old files left lying around
    utils.purge_old_files(args)

    # Run the command that the user specified
    return args.func(args)


if __name__ == '__main__':
    sys.exit(main())



# Author: Mark Blakeney, Feb 2024.
"Uninstall one, or more, or all applications."

from __future__ import annotations

from argparse import ArgumentParser, Namespace

from .. import utils

aliases = ['remove']


def _uninstall(args: Namespace, pkgname: str) -> str | None:
    "Uninstall given package"
    pkgname, vdir = utils.get_package_from_arg(pkgname, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    if not utils.rm_package(pkgname, args):
        return f'Failed to uninstall {pkgname}.'

    print(f'{pkgname} uninstalled.')
    return None


def init(parser: ArgumentParser) -> None:
    "Called to add command arguments to parser at init"
    parser.add_argument('-v', '--verbose', action='store_true', help='give more output')
    parser.add_argument('--all', action='store_true', help='uninstall ALL applications')
    parser.add_argument(
        '--skip',
        action='store_true',
        help='skip the specified applications when '
        'uninstalling all (only can be specified with --all)',
    )
    parser.add_argument(
        'package',
        nargs='*',
        help='application[s] to uninstall (or to skip for --all --skip)',
    )


def main(args: Namespace) -> str | None:
    "Called to action this command"
    for pkgname in utils.get_package_names(args):
        if error := _uninstall(args, pkgname):
            return error

    return None



# Author: Mark Blakeney, Feb 2024.
"Run pip with given arguments on virtual environment for the given application."

from __future__ import annotations

from argparse import ArgumentParser, Namespace

from .. import utils


def init(parser: ArgumentParser) -> None:
    "Called to add command arguments to parser at init"
    parser.add_argument('package', help='installed application name')
    parser.add_argument(
        'args', nargs='*', help='arguments to pass to uv pip, should start with "--".'
    )


def main(args: Namespace) -> str | None:
    "Called to action this command"
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    if not utils.piprun(vdir, args, args.args, quiet=True):
        return f'Error: failed to run pip for {pkgname}'
    return None



# Author: Mark Blakeney, Feb 2024.
"Common module containing shared functions"

from __future__ import annotations

import shlex
import subprocess
import sys
from typing import Sequence


def run(
    cmd: Sequence[str],
    *,
    capture: bool = False,
    quiet: bool = False,
    ignore_error=False,
) -> str | None:
    "Run given command"
    # Lazy evaluation of cmdstr
    cmdstr = None

    if capture:
        stdout = subprocess.PIPE
    else:
        stdout = None
        if not quiet:
            if not cmdstr:
                cmdstr = shlex.join(cmd)
            print(f'>>> Running {cmdstr}')
    try:
        res = subprocess.run(cmd, stdout=stdout, text=True)
    except Exception as e:
        if not ignore_error:
            if not cmdstr:
                cmdstr = shlex.join(cmd)
            print(f'{cmdstr} failed: {e}', file=sys.stderr)
        return None

    if res.returncode != 0:
        return None

    return (res.stdout and res.stdout.strip()) if capture else 'ok'



# Author: Mark Blakeney, Feb 2024.
"""
Install extra packages into an application.

Note the same --index-url is used as/if specified in the original install.
"""

from __future__ import annotations

from argparse import ArgumentParser, Namespace

from .. import utils


def init(parser: ArgumentParser) -> None:
    "Called to add command arguments to parser at init"
    parser.add_argument('-v', '--verbose', action='store_true', help='give more output')
    parser.add_argument('package', help='installed application name')
    parser.add_argument(
        'extras', nargs='+', help='extra package name[s] to inject/install'
    )


def main(args: Namespace) -> str | None:
    "Called to action this command"
    pkgname, vdir = utils.get_package_from_arg(args.package, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    data = utils.get_json(vdir, args) or {}
    url = data.get('url')
    pip_args = (
        'install --compile-bytecode -U'.split()
        + utils.make_args((args.verbose, '-v'), (url, ('-i', url)))
        + args.extras
    )
    if not utils.piprun(vdir, args, pip_args):
        return f'Error: failed to install "{args.extras}" to {pkgname}'

    return utils.add_or_remove_pkg(
        vdir, args, pkgname, args.extras, data=data, add=True
    )



# Author: Mark Blakeney, Feb 2024.
"Common module containing shared functions"

from __future__ import annotations

import json
import os
import shutil
import sys
from argparse import Namespace
from pathlib import Path
from typing import Iterable, Sequence

from .run import run

HOME = Path.home()


def subenvars(path: str, *, resolve: bool = False) -> Path:
    "Substitute environment variables in a path string"
    spath = Path(os.path.expandvars(path)).expanduser()
    return spath.resolve() if resolve else spath


def unexpanduser(path: str | Path) -> str:
    "Return path name, with $HOME replaced by ~ (opposite of Path.expanduser())"
    ppath = Path(path)

    if ppath.parts[: len(HOME.parts)] != HOME.parts:
        return str(path)

    return str(Path('~', *ppath.parts[len(HOME.parts) :]))


def get_json(vdir: Path, args: Namespace) -> dict | None:
    "Get JSON data for this virtual environment"
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open() as fp:
            return json.load(fp)
    except Exception:
        return None


def _set_json(vdir: Path, args: Namespace, data: dict) -> str | None:
    "Set JSON data for this virtual environment"
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open('w') as fp:
            json.dump(data, fp)
    except Exception as e:
        return str(e)

    return None


def piprun(vdir: Path, args: Namespace, cmd: list[str], **kargs) -> str | None:
    "Run given pip command in the virtual environment"
    # First element in cmd is the command, the rest are arguments. So
    # insert path to virtual environment after the command.
    cmd[1:1] = ['-p', str(vdir.resolve())]
    return run([args._uv, 'pip'] + cmd, **kargs)


def get_versions(
    vdir: Path, args: Namespace
) -> dict[str, tuple[str, str | None]] | None:
    "Return the versions of the packages in the virtual environment"
    if not (out := piprun(vdir, args, ['list', '-q'], capture=True)):
        return None

    found = False
    data: dict[str, tuple[str, str | None]] = {}
    for line in out.splitlines():
        if line.startswith('-'):
            found = True
        elif found:
            fields = line.split(maxsplit=2)
            if len(fields) == 2:
                data[fields[0]] = fields[1], None
            else:
                data[fields[0]] = fields[1], fields[2]

    return data


def make_args(*args: Sequence) -> list[str]:
    "Build a list of args based on (bool, arg1, [arg2]) sequences"
    retlist: list[str] = []
    for alist in args:
        if alist[0]:
            retlist.extend(alist[1:])

    return retlist


def vdir_bin(vdir: Path) -> Path:
    "Return the bin directory for the virtual environment"
    return vdir / 'bin'


def _load_record(rfile: Path) -> Iterable[str]:
    "Yield the executable names from a RECORD file"
    with rfile.open() as fp:
        for line in fp:
            line = line.strip()
            if line.startswith('../../../bin/'):
                yield Path(line.split(',', 1)[0]).name


def _link_app_files(
    vdir: Path, tgtdir: Path, pkgname: str, args: Namespace, include_deps: bool
) -> Iterable[str]:
    "Link app files from entry_points to tgtdir"
    vpath = vdir_bin(vdir)

    key = pkgname.replace('-', '_').lower() + '-'
    for efile in vdir.glob('**/*.dist-info/RECORD'):
        if include_deps or efile.parent.name.lower().startswith(key):
            for app in _load_record(efile):
                srcfile = vpath / app
                if (
                    srcfile.is_file()
                    and not srcfile.is_symlink()
                    and (srcfile.stat().st_mode & 0o111) == 0o111
                ):
                    tgtfile = tgtdir / app
                    if tgtfile.is_symlink():
                        tgtfile.unlink()

                    if args.verbose:
                        print(f'Linking "{srcfile}" -> "{tgtfile}"')

                    tgtfile.parent.mkdir(parents=True, exist_ok=True)
                    try:
                        tgtfile.symlink_to(srcfile)
                    except Exception as e:
                        print(f'Error: {e}', file=sys.stderr)
                        return

                    yield srcfile.name


def _link_all_files(srcdir: Path, tgtdir: Path, pat: str, verbose: bool) -> None:
    "Link files from srcdir to tgtdir"
    for srcfile in srcdir.glob(pat):
        tgtfile = tgtdir / srcfile.relative_to(srcdir)
        if tgtfile.is_symlink():
            tgtfile.unlink()

        if verbose:
            print(f'Linking "{srcfile}" -> "{tgtfile}"')

        tgtfile.parent.mkdir(parents=True, exist_ok=True)
        try:
            tgtfile.symlink_to(srcfile)
        except Exception as e:
            print(f'Error: {e}', file=sys.stderr)


def _unlink_all_files(vdir: Path, args: Namespace) -> None:
    "Unlink all link files"
    dirlist: list[tuple[Path, str]] = [(args._bin_dir, '*')]
    if args._man_dir.is_dir():
        dirlist.append((args._man_dir, '*/*'))

    for srcdir, pat in dirlist:
        if srcdir.is_dir():
            for file in srcdir.glob(pat):
                if file.is_symlink() and vdir in file.resolve().parents:
                    if args.verbose:
                        print(f'Removing link "{file}"')
                    file.unlink()


def make_links(
    vdir: Path, pkgname: str, args: Namespace, data: dict | None
) -> str | None:
    "[Re-]create app and man page links for package"

    vdir = vdir.resolve()

    # Unlink any existing links
    _unlink_all_files(vdir, args)

    # Link the package applications
    if not data and not (data := get_json(vdir, args)):
        return 'Error: No JSON data found.'

    include_deps = bool(
        args.include_deps if hasattr(args, 'include_deps') else data.get('deps')
    )

    apps = list(_link_app_files(vdir, args._bin_dir, pkgname, args, include_deps))

    # Link all the man pages
    if not args.no_man_pages:
        _link_all_files(vdir / 'share' / 'man', args._man_dir, '*/*', args.verbose)

    # Save the apps in the JSON data
    if not apps:
        return f'Error: {pkgname} has no executables to install.'

    if not (freeze := piprun(vdir, args, ['freeze'], capture=True)):
        return 'Error: Failed to fetch freeze list.'

    (vdir / args._freeze_file).write_text(freeze)
    data['apps'] = sorted(apps)
    return _set_json(vdir, args, data)


def rm_vdir(vdir: Path, args: Namespace) -> None:
    "Remove all links that point into the virtual environment"
    vdir = vdir.resolve()
    _unlink_all_files(vdir, args)

    # Remove the venv
    if vdir.exists():
        if args.verbose:
            print(f'Removing "{vdir}"')

        shutil.rmtree(vdir)


def _pkg_merge(inset: list[str], changeset: list[str], add: bool) -> Iterable[str]:
    "Merge a new list of package names/requirements"
    from packaging.requirements import Requirement

    inset_names = {Requirement(p).name: p for p in inset}
    changeset_names = {Requirement(p).name: p for p in changeset}

    if add:
        inset_names.update(changeset_names)
    else:
        for k in changeset_names:
            inset_names.pop(k, None)

    return inset_names.values()


def add_or_remove_pkg(
    vdir: Path,
    args: Namespace,
    pkgname: str,
    pkgs: list[str],
    *,
    add: bool,
    data: dict | None = None,
) -> str | None:
    "Record the addition/removal of a package into the virtual environment"
    if not data:
        data = get_json(vdir, args)
        if not data:
            return 'No JSON data found'

    newpkgs = _pkg_merge(data.get('injected', []), pkgs, add)

    if newpkgs:
        data['injected'] = sorted(newpkgs)
    else:
        data.pop('injected', None)

    return make_links(vdir, pkgname, args, data)


def rm_package(pkgname: str, args: Namespace) -> bool:
    "Remove this package"
    pdir = args._packages_dir / pkgname
    if not pdir.exists():
        return False

    vdir = pdir.resolve()

    if args.verbose:
        print(f'Removing link "{pdir}"')
    pdir.unlink()

    rm_vdir(vdir, args)
    return True


def get_all_pkg_venvs(args: Namespace) -> Iterable[tuple[Path, dict]]:
    "Return a list of all virtual environments and their JSON data"
    for pdir in sorted(args._packages_dir.iterdir()):
        if data := get_json(pdir, args):
            yield pdir, data


def _get_package_if_dir(name: str, args: Namespace) -> str | None:
    "Convert the given name if it corresponds to a package directory"
    if name not in {'.', '..'} and os.sep not in name:
        return name

    if not (namepath := Path(name).resolve()).is_dir():
        return None

    # Work out the package name from the current path. We look for
    # the closest parent (i.e. longest path) across all matching
    # application editpaths, unless we find an exact match then just
    # use that.
    candidates = {}
    for pdir, data in get_all_pkg_venvs(args):
        if data and (path := data.get('editpath')):
            # If we have an exact match then use it and ignore any
            # previous candidates.
            if (path := Path(path).expanduser()) == namepath:
                return pdir.name

            if path in namepath.parents:
                # This path has a candidate parent, so record it.
                candidates[len(path.parts)] = pdir.name

    return candidates[max(candidates)] if candidates else None


def get_package_from_arg(name: str, args: Namespace) -> tuple[str, Path | None]:
    "Return the package name + vdir corresponding to the given arg, if any"
    if not (pkg := _get_package_if_dir(name, args)):
        return name, None

    vdir = (args._packages_dir / pkg).resolve()
    return pkg, (vdir if vdir.exists() else None)


def _rm_path(path: Path) -> None:
    "Remove the given path"
    print(f'Purging stray "{path}"', file=sys.stderr)
    if path.is_symlink():
        path.unlink()
    elif path.is_dir():
        shutil.rmtree(path)
    elif path.exists():
        path.unlink()


def purge_old_files(args: Namespace) -> None:
    "Clean out any old virtual environments, packages, and executables"
    # Remove any packages that do not point to a dir in the venvs directory
    valids_venvs = set()
    for pkg in args._packages_dir.iterdir():
        vdir = pkg.resolve()
        if (
            vdir.parent != args._venvs_dir
            or not vdir.is_dir()
            or not vdir.name.isdigit()
        ):
            _rm_path(pkg)
        else:
            valids_venvs.add(vdir.name)

    # Remove any venvs that are not in the packages directory
    for vdir in args._venvs_dir.iterdir():
        if vdir.name not in valids_venvs:
            _rm_path(vdir)

    # Remove any executables that point to a non-existent path
    for exe in args._bin_dir.iterdir():
        rexe = exe.resolve()
        if args._venvs_dir in rexe.parents and not rexe.exists():
            _rm_path(exe)


def get_package_names(args: Namespace) -> list[str]:
    "Return a list of validated package names based on command line args"
    if args.all:
        if not args.skip and args.package:
            args.parser.error(
                'Can not specify package[s] with --all unless also specifying --skip.'
            )
    else:
        if args.skip:
            args.parser.error('--skip can only be specified with --all.')

        if not args.package:
            args.parser.error('Must specify at least one package, or --all.')

    given_names = set((_get_package_if_dir(p, args) or p) for p in args.package)
    all_names = set(f.name for f in args._packages_dir.iterdir() if f.is_dir())

    if unknown := given_names - all_names:
        s = 's' if len(unknown) > 1 else ''
        sys.exit(f'Error: Unknown package{s}: {", ".join(unknown)}')

    return sorted(all_names - given_names) if args.all else args.package


def get_python(args: Namespace) -> Path:
    "Return the python executable based on command line args"
    return subenvars(args.python) if args.python else args._pyexe


def version() -> str:
    "Return the version of this package"
    from importlib.metadata import version

    try:
        ver = version(Path(sys.argv[0]).stem)
    except Exception:
        ver = 'unknown'

    return ver



# Author: Mark Blakeney, Feb 2024.
"List applications installed by this tool."

from __future__ import annotations

import json
from argparse import ArgumentParser, Namespace

from .. import utils


def _show(value: str) -> str:
    "Show a string with quotes"
    return f'"{value}"' if isinstance(value, str) else value


def init(parser: ArgumentParser) -> None:
    "Called to add command arguments to parser at init"
    parser.add_argument('--json', action='store_true', help='output json instead')
    parser.add_argument(
        '-v',
        '--venv',
        action='store_true',
        help='also show the virtual environment dir/number',
    )
    parser.add_argument('package', nargs='*', help='list the given application[s] only')


def main(args: Namespace) -> str | None:
    "Called to action this command"
    if args.package:
        pkgs = [utils.get_package_from_arg(p, args) for p in args.package]
    else:
        pkgs = sorted((p.name, p) for p in args._packages_dir.iterdir())

    json_out = {}
    for pkgname, vdir in pkgs:
        if not vdir:
            return f'Application {pkgname} is not installed.'

        if data := utils.get_json(vdir, args):
            data.pop('name', None)
            if args.venv:
                data['venv'] = int(vdir.resolve().name)
            if args.json:
                json_out[pkgname] = data
            else:
                d = ', '.join(f'{k}={_show(data[k])}' for k in sorted(data))
                print(f'{pkgname}: {d}')

    if args.json:
        print(json.dumps(json_out, indent=2))

    return None



# Author: Mark Blakeney, Feb 2024.
"Reinstall one, or more, or all applications."

from __future__ import annotations

import shutil
import tempfile
from argparse import ArgumentParser, Namespace
from copy import copy
from pathlib import Path

from .. import utils
from ..run import run


def _reinstall(args: Namespace, pkgname: str, venv_args: list[str]) -> str | None:
    "Reinstall given application"
    pkgname, vdir = utils.get_package_from_arg(pkgname, args)
    if not vdir:
        return f'Application {pkgname} is not installed.'

    print(f'Reinstalling {pkgname} ..')
    pip_args = 'sync --compile-bytecode --reinstall'.split() + utils.make_args(
        (args.verbose, '-v')
    )

    data = utils.get_json(vdir, args) or {}
    if url := data.get('url'):
        pip_args.extend(['-i', url])

    # Use explicit python (or reset) if given, else use what was
    # explicitly specified in original install, else use default
    # python
    nargs = copy(args)
    if args.reset_python:
        data.pop('python', None)
    elif args.python:
        data['python'] = utils.unexpanduser(args.python)
    else:
        nargs.python = data.get('python')

        # Update python path if still in expanded (old) format
        if nargs.python:
            data['python'] = utils.unexpanduser(nargs.python)

    venv_args.extend(['-p', str(utils.get_python(nargs))])

    if args.system_site_packages:
        data['sys'] = True
    elif args.no_system_site_packages:
        data.pop('sys', None)

    if data.get('sys'):
        venv_args.append('--system-site-packages')

    # Update editpath if still in expanded (old) format
    if editpath := data.get('editpath'):
        data['editpath'] = utils.unexpanduser(editpath)

    with tempfile.TemporaryDirectory() as tdir:
        tfile = Path(tdir, args._freeze_file)
        shutil.copyfile(vdir / args._freeze_file, tfile)

        # Recreate the vdir
        if not run(venv_args + [str(vdir)]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to recreate {vdir} for {pkgname}.'

        if not utils.piprun(vdir, args, pip_args + [str(tfile)]):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to resync {pkgname}'

    if err := utils.make_links(vdir, pkgname, args, data):
        return err

    print(f'{pkgname} reinstalled.')
    return None


def init(parser: ArgumentParser) -> None:
    "Called to add command arguments to parser at init"
    xgroup = parser.add_mutually_exclusive_group()
    xgroup.add_argument(
        '-p', '--python', help='specify explicit python executable path'
    )
    xgroup.add_argument(
        '--reset-python',
        action='store_true',
        help='reset any explicit python path to default python',
    )
    ygroup = parser.add_mutually_exclusive_group()
    ygroup.add_argument(
        '--system-site-packages',
        action='store_true',
        help='allow venv access to system packages, '
        'overrides the per-application setting',
    )
    ygroup.add_argument(
        '--no-system-site-packages',
        action='store_true',
        help='remove venv access to system packages, '
        'overrides the per-application setting',
    )
    parser.add_argument('-v', '--verbose', action='store_true', help='give more output')
    parser.add_argument('--all', action='store_true', help='reinstall ALL applications')
    parser.add_argument(
        '--skip',
        action='store_true',
        help='skip the specified applications when '
        'reinstalling all (only can be specified with --all)',
    )
    parser.add_argument(
        'package',
        nargs='*',
        help='application[s] to reinstall (or to skip for --all --skip)',
    )


def main(args: Namespace) -> str | None:
    "Called to action this command"
    venv_args = [args._uv, 'venv'] + utils.make_args(
        (args.verbose, '-v'), (not args.verbose, '-q')
    )
    for pkgname in utils.get_package_names(args):
        if error := _reinstall(args, pkgname, venv_args.copy()):
            return error

    return None
